#include <dune/common/fmatrix.hh>

#include <dune/istl/matrix.hh>
#include <dune/istl/matrixindexset.hh>
#include <dune/istl/bcrsmatrix.hh>


#include <dune/geometry/quadraturerules.hh>

#include "elasticityhelpers.hh"


template <class GridType, class TrialLocalFE, class AnsatzLocalFE, int polOrd>
void NeoHookeLocalAssembler<GridType,TrialLocalFE,AnsatzLocalFE, polOrd>::
assemble(const ElementPointer& element,
		 const Dune::BlockVector<Dune::FieldVector<double,dim> >& localSolution,
		 LocalMatrix& localMatrix, 
		 Dune::BlockVector<Dune::FieldVector<double,dim> >& localRhs, 
		 const TrialLocalFE& tFE, const AnsatzLocalFE& aFE) const
{
    typedef typename TrialLocalFE::Traits::LocalBasisType::Traits::JacobianType JacobianType;
    
    int rows = tFE.localBasis().size();
    int cols = aFE.localBasis().size();

    localMatrix.setSize(rows,cols);
    localMatrix = 0.0;
	
    localRhs.resize(rows);
    localRhs = 0;

    /* ndof is the number of vectors of the element */
    int ndof = tFE.localBasis().size();
    int ncomp = dim;

    // Get quadrature rule
    const Dune::QuadratureRule<double, dim>& quad = Dune::QuadratureRules<double, dim>::rule(element->type(), polOrd);

    // store gradients of shape functions and base functions
    std::vector<JacobianType> referenceGradients(tFE.localBasis().size());
    std::vector<Dune::FieldVector<double,dim> > gradients(tFE.localBasis().size());
    
    /* Loop over all integration points */
    for (int ip=0; ip<quad.size(); ip++) {
        
        // Local position of the quadrature point
        const Dune::FieldVector<double,dim> quadPos = quad[ip].position();

        // calc Jacobian inverse before integration element is evaluated 
        const Dune::FieldMatrix<double,dim,dim>& inv = element->geometry().jacobianInverseTransposed(quadPos);
        const double integrationElement = element->geometry().integrationElement(quadPos);
        
        /* Compute the weight of the current integration point */
        double weight = quad[ip].weight() * integrationElement;
        
        /**********************************************/
        /* compute gradients of the shape functions   */
        /**********************************************/
        // get gradients of shape functions
        tFE.localBasis().evaluateJacobian(quadPos, referenceGradients);
                
        // compute gradients of base functions
        for (int i=0; i<gradients.size(); ++i) {

        	// transform gradients
        	gradients[i] = 0.0;
        	inv.umv(referenceGradients[i][0], gradients[i]);

        }
        
        /****************************************************/
        /* The deformation gradients of the shape functions */
        /****************************************************/
        
        Dune::FieldMatrix<double,dim,dim> nablaPhi[100*dim];
        for (int dof=0; dof<ndof; dof++)
            for (int comp=0; comp<ncomp; comp++)
                for (int i=0; i<dim; i++)
                    for (int j=0; j<dim; j++)
                        nablaPhi[dof*ncomp+comp][i][j] = (i==comp) ? gradients[dof][j] : 0; 
      
        /****************************************************/
        // The deformation gradient of the deformation
        // in formulas: F(i,j) = $\partial \phi_i / \partial x_j$
        // or F(i,j) = $\partial u_i / \partial x_j$
        /****************************************************/
                        Dune::FieldMatrix<double, dim, dim> parDerU;
        for (int i=0; i<ncomp; i++) {
            
            for (int j=0; j<dim; j++) {
                
                parDerU[i][j] = 0;
                
                for (int k=0; k<ndof; k++)
                    parDerU[i][j] += localSolution[k][i]*gradients[k][j];
                
            }

        }

        // J = det (I + \nabla U)
        const double J = Dune::Elasticity::det_val(parDerU);

        // ///////////////////////////////////////////////////////
        //   The material moduli C_ijkl
        // ///////////////////////////////////////////////////////
        Dune::FieldMatrix<double,dim,dim> parJparU;

        parJparU[0][0] = Dune::Elasticity::det_du_tmpl<0,0>(parDerU);
        parJparU[0][1] = Dune:: Elasticity::det_du_tmpl<0,1>(parDerU);
        parJparU[1][0] = Dune::Elasticity::det_du_tmpl<1,0>(parDerU);
        parJparU[1][1] = Dune::Elasticity::det_du_tmpl<1,1>(parDerU);
        if (dim==3) {
            parJparU[0][2] = Dune::Elasticity::det_du_tmpl<0,2>(parDerU);
            parJparU[1][2] = Dune::Elasticity::det_du_tmpl<1,2>(parDerU);
            parJparU[2][0] = Dune::Elasticity::det_du_tmpl<2,0>(parDerU);
            parJparU[2][1] = Dune::Elasticity::det_du_tmpl<2,1>(parDerU);
            parJparU[2][2] = Dune::Elasticity::det_du_tmpl<2,2>(parDerU);
        }
        
        double C[dim][dim][dim][dim];

        for (int i=0; i<dim; i++) {
            
            for (int k=0; k<dim; k++) {

                for (int j=0; j<dim; j++) {
                
                    for (int l=0; l<dim; l++) {

                        double tr_E_dudu = 0;
                        for (int m=0; m<dim; m++)
                            tr_E_dudu += Dune::Elasticity::e_dudu(m,m,i,j,k,l);
                            
#define LAURSEN
                        // Laursen
#ifdef LAURSEN
                        C[i][j][k][l] = lambda_/2 * (parJparU[i][j]*parJparU[k][l] + J*Dune::Elasticity::det_dudu_tmpl(parDerU,i,j,k,l));
                        C[i][j][k][l] += -(lambda_/2 + mu_)/J/J * (J*Dune::Elasticity::det_dudu_tmpl(parDerU,i,j,k,l) - parJparU[i][j]*parJparU[k][l]);
                        C[i][j][k][l] += mu_ * tr_E_dudu;
#else
                        // Fischer / Wriggers
                        C[i][j][k][l] = lambda_*(parJparU[i][j]*parJparU[k][l] + (J-1)*Dune::Elasticity::det_dudu_tmpl(parDerU,i,j,k,l));
                        C[i][j][k][l] += -(2*mu_)/J/J * (J*Dune::Elasticity::det_dudu_tmpl(parDerU,i,j,k,l) - parJparU[i][j]*parJparU[k][l]);
                        C[i][j][k][l] += mu_ * tr_E_dudu;
#endif
                    }

                }

            }

        }

        // /////////////////////////////////////////////////////////////////////
        //   The gradient of the energy functional (for the right hand side)
        // /////////////////////////////////////////////////////////////////////

        double trE_du[dim][dim];
        for (int i=0; i<dim; ++i)
            for (int j=0; j<dim; ++j) {
                trE_du[i][j] = 0;
                for (int k=0; k<dim; k++)
                    trE_du[i][j] += Dune::Elasticity::E_du(parDerU,k,k,i,j);
            }

        Dune::FieldMatrix<double,dim,dim> fu(0);

        for (int i=0; i<dim; i++) {

            for (int j=0; j<dim; j++) {

                // Laursen
#ifdef LAURSEN
                fu[i][j]  = J * lambda_ / 2 * parJparU[i][j];
                fu[i][j] -= (lambda_/2 + mu_)/J * parJparU[i][j];
                fu[i][j] += mu_ * trE_du[i][j];
#else
                // Fischer
                fu[i][j]  = lambda_ * (J-1) * parJparU[i][j];
                fu[i][j] -= 2* mu_ /J * parJparU[i][j];
                fu[i][j] += mu_ * trE_du[i][j];
#endif
            }

        }

        //std::cout << "fu: " << std::endl << fu << std::endl;
        // loop over all entries of the element stiffness matrix
        // and add the contribution of the current quadrature point
        for (int i=0; i<ndof*ncomp; i++) {

            for (int j=0; j<ndof*ncomp; j++) {
                
                // Compute \nabla \phi_i * C * \nabla \phi_j

                // First tensor product:  
            	Dune::FieldMatrix<double,dim,dim> tmp(0);
                for (int k=0; k<dim; k++)
                    for (int l=0; l<dim; l++)
                        for (int m=0; m<dim; m++)
                            for (int n=0; n<dim; n++)
                                tmp[k][l] += C[k][l][m][n]*nablaPhi[j][m][n];
                
                // Second inner tensor product.
                for (int k=0; k<dim; k++)
                    for (int l=0; l<dim; l++) {
                    	localMatrix[i/dim][j/dim][i%dim][j%dim] += nablaPhi[i][k][l] * tmp[k][l] * weight;
                    }
                
            }
            

            /* add the nonlinear part to the right hand side */
            for (int k=0; k<dim; k++)
                for (int l=0; l<dim; l++)
                    localRhs[i/dim][i%dim] -= nablaPhi[i][k][l] * fu[k][l] * weight;
            
        }
        
    }
        
#if 0
    static int eleme = 0;
    printf("********* Element %d **********\n", eleme++);

    for (int row=0; row<matSize; row++) {
        
        for (int rcomp=0; rcomp<dim; rcomp++) {
            
            for (int col=0; col<matSize; col++) {
                
                for (int ccomp=0; ccomp<dim; ccomp++)
                    std::cout << mat[row][col][rcomp][ccomp] << "  ";
                
                std::cout << "    ";
                
            }
            
            std::cout << std::endl;
            
        }
        
        std::cout << std::endl;
        
    }
    //exit(0);
#endif
    
}


template <class TrialBasis,class AnsatzBasis,int polOrd>
void Dune::NeoHookeAssembler<TrialBasis, AnsatzBasis, polOrd>::
assembleProblem(BCRSMatrix<FieldMatrix<double,dim,dim> >& globalMatrix,
				const BlockVector<FieldVector<double, dim> >& sol,
                BlockVector<FieldVector<double, dim> >& rhs)
{
	//use method from the base class to compute indices of the matrix
	int rows = this->tBasis_.size();
	int cols = this->aBasis_.size();

	Dune::MatrixIndexSet indices(rows, cols);
	    
	// Create local assembler
	NeoHookeLocalAssembler<typename GridView::Grid,typename TrialBasis::LocalFiniteElement,typename AnsatzBasis::LocalFiniteElement, polOrd> localAssembler(lambda_, mu_);
	      
	addIndices(localAssembler, indices, false);
	
	indices.exportIdx(globalMatrix);
	globalMatrix=0.0;

	typedef typename NeoHookeLocalAssembler<typename GridView::Grid,typename TrialBasis::LocalFiniteElement,typename AnsatzBasis::LocalFiniteElement, polOrd>::LocalMatrix LocalMatrix;
	    

	ElementIterator it = this->tBasis_.getGridView().template begin<0>();
	ElementIterator endIt = this->tBasis_.getGridView().template end<0>();
      
    for( ; it != endIt; ++it ) {
        
    	// get shape functions
    	const typename TrialBasis::LocalFiniteElement& tFE = this->tBasis_.getLocalFiniteElement(*it);
    	const typename AnsatzBasis::LocalFiniteElement& aFE = this->aBasis_.getLocalFiniteElement(*it); 
    	    	    	
    	LocalMatrix localA(tFE.localBasis().size(), aFE.localBasis().size());
    	    	
    	// Extract local solution
    	BlockVector<FieldVector<double, dim> > localSolution(aFE.localBasis().size()), localRhs(tFE.localBasis().size());
    	        
    	for (int i=0; i<aFE.localBasis().size(); i++)
    		localSolution[i] = sol[this->aBasis_.index(*it,i)];
        
    	// setup matrix 
        localAssembler.assemble(it, localSolution, localA, localRhs, tFE, aFE);
        
        // Add element matrix to global stiffness matrix
        for(int i=0; i<tFE.localBasis().size(); i++) { 
        	
        	int rowIndex = this->tBasis_.index(*it, i);
        	
        	for (int j=0; j<aFE.localBasis().size(); j++ ) {
        		
        		int colIndex = this->aBasis_.index(*it, j);
        		globalMatrix[rowIndex][colIndex] += localA[i][j];
        		
        	}
        }
        
        // Add elements of local rhs to global rhs
        for (int i=0; i<tFE.localBasis().size(); i++)
        	rhs[this->tBasis_.index(*it, i)] += localRhs[i];
    }
    
}

template <class TrialBasis,class AnsatzBasis,int polOrd>
double Dune::NeoHookeAssembler<TrialBasis,AnsatzBasis, polOrd>::
computeEnergy(const BlockVector<FieldVector<double, dim> >& sol) const
{

    double energy = 0;
    
    if (sol.size()!=this->aBasis_.size())
        DUNE_THROW(Exception, "Solution vector doesn't match the grid!");

    ElementIterator it = this->tBasis_.getGridView().template begin<0>();
    ElementIterator endIt = this->tBasis_.getGridView().template end<0>();

    // Loop over all elements
    for (; it!=endIt; ++it) {

        // Extract local solution on this element
    	const typename AnsatzBasis::LocalFiniteElement& localFE = this->aBasis_.getLocalFiniteElement(*it);
    	        
    	FieldVector<double, dim> localSolution[localFE.localBasis().size()];
        
        // In our implementation
        // The current solution is a deformation, not a position!
    	for (int i=0; i<localFE.localBasis().size(); i++)
    		localSolution[i] = sol[this->aBasis_.index(*it, i)];

        // Get quadrature rule
        const QuadratureRule<double, dim>& quad = QuadratureRules<double, dim>::rule(it->type(), polOrd);        

        for (int pt=0; pt<quad.size(); pt++) {

            // Local position of the quadrature point
            const FieldVector<double,dim>& quadPos = quad[pt].position();
            
            const MatrixBlock& inv = it->geometry().jacobianInverseTransposed(quadPos);
            const double integrationElement = it->geometry().integrationElement(quadPos);
        
            double weight = quad[pt].weight() * integrationElement;
            
            // ///////////////////////////////////////
            //   Compute deformation gradient
            // ///////////////////////////////////////
            std::vector<FieldMatrix<double,1, dim> > shape_grads;
            localFE.localBasis().evaluateJacobian(quadPos,shape_grads);
            
            for (int dof=0; dof<localFE.localBasis().size(); dof++) {
                       
            	//std::cout << "Gradient " << dof << ": " << shape_grads[dof] << std::endl;
            	
            	// multiply with jacobian inverse 
            	FieldVector<double, dim> tmp(0);
            	inv.umv(shape_grads[dof][0], tmp);
            	shape_grads[dof][0] = tmp;
            	//std::cout << "Gradient " << dof << ": " << shape_grads[dof] << std::endl;
            }

            // ///////////////////////////////////////////////
            //   Compute the deformation gradient
            // ///////////////////////////////////////////////
            FieldMatrix<double, dim, dim> defGrad;
            for (int i=0; i<dim; i++) {
                
                for (int j=0; j<dim; j++) {
                    
                    defGrad[i][j] = 0;
                    
                    for (int k=0; k<localFE.localBasis().size(); k++)
                        defGrad[i][j] += localSolution[k][i]*shape_grads[k][0][j];
                    
                }
                
            }

            // /////////////////////////////////////////////
            //   The trace of the nonlinear strain tensor
            // /////////////////////////////////////////////
            FieldMatrix<double,dim,dim> E;
            for (int i=0; i<dim; i++)
                for (int j=0; j<dim; j++)
                    E[i][j] = Elasticity::E_val(defGrad, i, j);

            double trE = 0;
            for (int i=0; i<dim; i++)
                trE += E[i][i];
            
            // ///////////////////////////////////////
            //   Sum up result
            // ///////////////////////////////////////

            const double J = Elasticity::det_val(defGrad);//defGrad.determinant();

//             printf("weight: %g, lambda: %g  mu: %g  J: %g,   trE: %g, adding %g\n", 
//                    quad[pt].weight(), lambda_, mu_, J, trE,
//                    weight * (lambda_ * (J*J-1)/4 - (lambda_/2 + mu_ ) * std::log(J) + mu_*trE));
            // Laursen
#ifdef LAURSEN
            energy += weight * (lambda_ * (J*J-1)/4 - (lambda_/2 + mu_ ) * std::log(J) + mu_*trE);
#else
            // Fischer
            energy += weight * (lambda_ * (J-1)*(J-1)/2 - 2*mu_ * std::log(J) + mu_*trE);
#endif
        }

    }

    return energy;

}
