#ifndef DUNE_OGDEN_ELASTICITY_OPERATOR_HH
#define DUNE_OGDEN_ELASTICITY_OPERATOR_HH

#include <dune/common/fmatrix.hh>

#include <dune/istl/bcrsmatrix.hh>
#include <dune/istl/matrixindexset.hh>
#include <dune/istl/matrix.hh>

#include <dune/fufem/assemblers/localassembler.hh>
#include <dune/fufem/assemblers/localoperatorassembler.hh>
#include <dune/fufem/assemblers/operatorassembler.hh>

  /**
   * @brief compute local stiffness matrix for the Ogden elasticity operator
   *
   */


  //! A class for computing local stiffness matrices
  /*! A class for computing local stiffness matrix for the 
	linear elasticity equation

        - div \sigma = f in Omega

        u = g on Gamma1; j*n = J on Gamma2.

	Uses conforming finite elements with the Lagrange shape functions.
	It should work for all dimensions and element types.
	All the numbering is with respect to the reference element and the
	Lagrange shape functions

	\tparam GridType  	   a DUNE grid type
	\tparam TrialLocalFE   a local finite element of the trial space
	\tparam AnsatzLocalFE  a local finite element of the ansatz space
  */
template <class GridType, class TrialLocalFE, class AnsatzLocalFE>
  class OgdenMaterialLocalStiffness 
       :public LocalOperatorAssembler < GridType, TrialLocalFE, AnsatzLocalFE, Dune::FieldMatrix<double,GridType::dimension,GridType::dimension> >
  {
 
  public:	   
	  typedef typename Dune::FieldMatrix<double,GridType::dimension,GridType::dimension> T;
	  typedef typename LocalAssembler < GridType, TrialLocalFE, AnsatzLocalFE, T >::Element Element;
	  typedef typename LocalAssembler < GridType, TrialLocalFE, AnsatzLocalFE,T >::BoolMatrix BoolMatrix;
	  typedef typename LocalAssembler < GridType, TrialLocalFE, AnsatzLocalFE,T >::LocalMatrix LocalMatrix;
	  
	  // some other sizes
	  enum {dim=GridType::dimension};

	  // define the number of components of your system, this is used outside
	  // to allocate the correct size of (dense) blocks with a FieldMatrix
	  enum {m=dim};
	  
	  // /////////////////////////////////
	  //   The material parameters
	  // /////////////////////////////////
        
      /** \brief Material constants */
        double lambda_;
        double mu_;

        /** \brief The Ogden constant */
        double d_;

      //! Default Constructor
      OgdenMaterialLocalStiffness ()
      {}

	//! Constructor
      OgdenMaterialLocalStiffness (double lambda, double mu, double d)
          : lambda_(lambda), mu_(mu), d_(d)
      {}

      /** \brief Set material parameters */
      void setEandNu(double E, double nu, double d)
      {
          lambda_ = E*nu / ((1+nu)*(1-2*nu));
          mu_     = E / (2*(1+nu));
          d_      = d;
      }
      
      void indices(const Element& element, BoolMatrix& isNonZero, const TrialLocalFE& tFE, const AnsatzLocalFE& aFE) const
          {
              isNonZero = true;
          }
      
      void assemble(const Element& element, 
    		  		const Dune::BlockVector<Dune::FieldVector<double,dim> >& localSolution,
    		  		LocalMatrix& localMatrix, 
    		  		const TrialLocalFE& tFE, const AnsatzLocalFE& aFE) const
      {
    	  Dune::BlockVector<Dune::FieldVector<double,dim> > dummy;
    	  dummy=0;
    	  
    	  assemble(element,localSolution,localMatrix,dummy,tFE,aFE);
    	           
      }   
      //this method is used by the OgdenAssembler to assemble the matrix and the right hand side 
      void assemble(const Element& element, 
    		  		const Dune::BlockVector<Dune::FieldVector<double,dim> >& localSolution,
    		  		LocalMatrix& localMatrix, Dune::BlockVector<Dune::FieldVector<double,dim> >& localRhs, 
    		  		const TrialLocalFE& tFE, const AnsatzLocalFE& aFE) const;
      // This method is never actually called, it's just here to implement all pure virtual functions
      void assemble(const Element& element, LocalMatrix& localMatrix,
                const TrialLocalFE& tFE, const AnsatzLocalFE& aFE) const
      {
          // dummy
          Dune::BlockVector<Dune::FieldVector<double,dim> > localSol(tFE.localBasis().size());
          localSol = 0;

          assemble(element,localSol,localMatrix,tFE, aFE);
      }
  };


namespace Dune 
{

    /** \brief The FEM operator for elasticity with an Ogden material
     */
template <class TrialBasis, class AnsatzBasis>
    class OgdenAssembler :
    public OperatorAssembler<TrialBasis,AnsatzBasis>
    {
    public:
         
     	typedef typename TrialBasis::GridView GridView;  	 
     	typedef typename GridView::template Codim<0>::Iterator ElementIterator;

     	//! The grid's dimension
     	enum { dim = GridView::dimension };
        
        enum { elementOrder = 1 };

        /** \todo Choose correct quadrature order for the different elements */
        enum { polOrd = 3};

        //!
        typedef FieldMatrix<double, dim, dim> MatrixBlock;
              
        /** \brief Material constants */
        double lambda_;
        double mu_;

        /** \brief The Ogden constant */
        double d_;

        //! ???
         OgdenAssembler(const TrialBasis& tBasis, const AnsatzBasis& aBasis) : 
         	OperatorAssembler<TrialBasis,AnsatzBasis>(tBasis,aBasis)
         { 
             d_ = 1e5;
             lambda_ = 144230;
             mu_     = 96153;
         }
                
        /** \brief Set the material parameters
            \param E Young's modulus
            \param nu Poisson ratio
        */
        void setEandNu(double E, double nu) {
            lambda_ = E*nu / ((1+nu)*(1-2*nu));
            mu_     = E / (2*(1+nu));
        }
              
        /** \brief Assemble the tangent stiffness matrix and the right hand side
         */
        void assembleProblem(BCRSMatrix<FieldMatrix<double,dim,dim> >& globalMatrix,
        		             const BlockVector<FieldVector<double, dim> >& sol,
                             BlockVector<FieldVector<double, dim> >& rhs);
        /* Not implemented and not used?
        /** \brief Compute the element tangent stiffness matrix  /
        template <class MatrixType>
        void getLocalMatrix( EntityType &entity, 
                             const BlockVector<FieldVector<double, dim> >& localSolution, 
                             const int matSize, MatrixType& mat,
                             BlockVector<FieldVector<double, dim> >& localRhs) const;
         */
        /** \brief Compute the Ogden energy of a deformation state */
        double computeEnergy(const BlockVector<FieldVector<double, dim> >& sol) const;
        
        
        
        
        
        /**********************************************************************/
        /*                                                                    */
        /* Assembly of large strain elasticity with Ogden material            */
        /*                                                                    */
        /**********************************************************************/
        
        /* The stuff from here up to AssembleOgdenMatrix is taken from
           Martin Weiser's Code */
        
        
        static double delta(int i, int j) { return i==j? 1.0: 0.0; }
        
        /*----------------------------------------------------------------------------*/
        
        
        /** \brief Compute the geometrically nonlinear strain tensor 
            \param u The deformation gradient: \f$\partial u_i / \partial x_j \f$
            \param i,j The entry we're interested in
            \return The entry \f$ E_{i,j} \f$ of the nonlinear strain tensor
        */
        template <int dim>
        static double E_val(const FieldMatrix<double,dim,dim>& u, int i, int j) 
        {
            double result = u[i][j] + u[j][i];
            
            for (int k=0; k<dim; ++k)
                result += u[k][i]*u[k][j];  
            return 0.5*result;
        }

        template <int dim>
        static double E_du(const FieldMatrix<double,dim,dim>& u, int i, int j, int p, int q) 
        {
            double result = delta(i,p)*delta(j,q) + delta(j,p)*delta(i,q);
            
            result += ( delta(i,q)*u[p][j] + u[p][i]*delta(j,q) );
            return 0.5*result;
        }
        
        // The Kronecker delta function
        template <int i, int j>
        struct Delta {
            static int const delta = 0;
        };
        
        template <int i>
        struct Delta<i,i> {
            static int const delta = 1;
        };
        
        template <int i, int j, int p, int q, int dim>
        static inline double e_du(const FieldMatrix<double,dim,dim>& u) 
        {
            double result = Delta<i,p>::delta*Delta<j,q>::delta
                + Delta<j,p>::delta*Delta<i,q>::delta;
            
            result += Delta<i,q>::delta*u[p][j] + u[p][i]*Delta<j,q>::delta;
            return 0.5*result;
        }
        
        template <int i, int j, int p, int q, int r, int s>
        static inline double e_dudu() 
        {
            return 0.5 * Delta<p,r>::delta
                * ( Delta<i,q>::delta*Delta<j,s>::delta + Delta<i,s>::delta*Delta<j,q>::delta );
        }
        
        
        /** \brief Computes the value of det(I+u') 
            \param u The deformation gradient: \f$\partial u_i / \partial x_j \f$
        */
        static double det_val(const FieldMatrix<double,3,3>& u) {
            return (1+u[0][0])*(1+u[1][1])*(1+u[2][2])
                + u[0][1]*u[1][2]*u[2][0]
                + u[0][2]*u[1][0]*u[2][1]
                - u[2][0]*(1+u[1][1])*u[0][2]
                - u[2][1]*u[1][2]*(1+u[0][0])
                - (1+u[2][2])*u[1][0]*u[0][1];
        }

        static double det_val(const FieldMatrix<double,2,2>& u) {
            return (1+u[0][0])*(1+u[1][1]) - u[1][0]*u[0][1];
        }
        
        /*----------------------------------------------------------------------------*/
        /* The Gamma term in the Ogden material. */
        
        /** \brief The penalty term \f$ \Gamma \f$ */
        static double Gamma(double x) {
            return x*x - log(x);
        }
        
        /** \brief First derivative of the penalty term \f$ \Gamma \f$ */
        static double Gamma_x(double x) {
            return 2*x - 1/x;
        }

        /** \brief Second derivative of the penalty term \f$ \Gamma \f$ */
        static double Gamma_xx(double x) {
            return 2 + 1/(x*x);
        }
        
        /*----------------------------------------------------------------------------*/
        
        /* Stored energy function of an Ogden-type hyperelastic material defined as
           W = a*tr(E) + b*tr(E)^2 + c*tr(E*E) + d*Gamma(det(I+u'))
           with a = -d*Gamma_x(1)
           b = (lambda - d*(Gamma_x(1)+Gamma_xx(1)))/2
           c = my + d*Gamma_x(1)
           such that for small deformations a St.Venant-Kirchhoff-Material is
           approximated with Parameters lambda and mu. For d=0, this linear
           material law is exactly reproduced.
        */
        
        // \parder {det I + u}{ u_pq }
        template <int p, int qq>
        static double det_du_tmpl(const FieldMatrix<double,3,3>& u) {
            int const q = qq+3;
            
            return (Delta<(p+1)%3,(q+1)%3>::delta + u[(p+1)%3][(q+1)%3])
                * (Delta<(p+2)%3,(q+2)%3>::delta+u[(p+2)%3][(q+2)%3]) 
                - (Delta<(p+1)%3,(q-1)%3>::delta+u[(p+1)%3][(q-1)%3])
                * (Delta<(p+2)%3,(q-2)%3>::delta+u[(p+2)%3][(q-2)%3]);
        }

        // \parder {det I + u}{ u_pq }
        template <int p, int q>
        static double det_du_tmpl(const FieldMatrix<double,2,2>& u) {
            return Delta<p,q>::delta * (1+u[(p+1)%2][(q+1)%2]) 
                - Delta<p,(q+1)%2>::delta * u[(p+1)%2][(q+1)%2];

            // p = 0, q = 0;  (1+u[1][1]) 
            // p = 0, q = 1; -u[1][0] 
            // p = 1, q = 0; -u[0][1]
            // p = 1, q = 1;  (1+u[0][0])
            //
        }

        template <int p, int qq, int r, int s>
        static double det_dudu_tmpl(const FieldMatrix<double,3,3>& u) {
            int const q = qq+3;
            
            return
                  Delta<(p+1)%3,r>::delta*Delta<(q+1)%3,s>::delta * (Delta<(p+2)%3,(q+2)%3>::delta+u[(p+2)%3][(q+2)%3])
                + Delta<(p+2)%3,r>::delta*Delta<(q+2)%3,s>::delta * (Delta<(p+1)%3,(q+1)%3>::delta+u[(p+1)%3][(q+1)%3])
                - Delta<(p+1)%3,r>::delta*Delta<(q-1)%3,s>::delta * (Delta<(p+2)%3,(q-2)%3>::delta+u[(p+2)%3][(q-2)%3])
                - Delta<(p+2)%3,r>::delta*Delta<(q-2)%3,s>::delta * (Delta<(p+1)%3,(q-1)%3>::delta+u[(p+1)%3][(q-1)%3]);
        }

        // parder det (I + u) partial u_pq \partial u_rs
        template <int p, int q, int r, int s>
        static double det_dudu_tmpl(const FieldMatrix<double,2,2>& u) {
            return delta(p,q)*delta(r,(p+1)%2)*delta(s,(q+1)%2)
                - delta(p,(q+1)%2)*delta(r,(p+1)%2)*delta(s,(q+1)%2);
        }

        /*----------------------------------------------------------------------------*/
        
        template <int p, int q>
        static double compute_fu(const FieldMatrix<double,3,3>& u, 
                                 double const E[3][3], double trE, double dGamma,
                                 double a, double b, double c, double d) 
        {
            
            double trE_dupq = e_du<0,0,p,q>(u)+e_du<1,1,p,q>(u)+e_du<2,2,p,q>(u);
            
            double tmp = 0;
            tmp += E[0][0]*e_du<0,0,p,q>(u);
            tmp += E[0][1]*e_du<0,1,p,q>(u);
            tmp += E[0][2]*e_du<0,2,p,q>(u);
            tmp += E[1][0]*e_du<1,0,p,q>(u);
            tmp += E[1][1]*e_du<1,1,p,q>(u);
            tmp += E[1][2]*e_du<1,2,p,q>(u);
            tmp += E[2][0]*e_du<2,0,p,q>(u);
            tmp += E[2][1]*e_du<2,1,p,q>(u);
            tmp += E[2][2]*e_du<2,2,p,q>(u);
            
            return a*trE_dupq + 2*b*trE*trE_dupq + 2*c*tmp + d*dGamma*det_du_tmpl<p,q>(u);
        }

        template <int p, int q>
        static double compute_fu(const FieldMatrix<double,2,2>& u, 
                                 double const E[2][2], double trE, double dGamma,
                                 double a, double b, double c, double d) 
        {
            
            double trE_dupq = e_du<0,0,p,q>(u)+e_du<1,1,p,q>(u);
            
            double tmp = 0;
            tmp += E[0][0]*e_du<0,0,p,q>(u);
            tmp += E[0][1]*e_du<0,1,p,q>(u);
            tmp += E[1][0]*e_du<1,0,p,q>(u);
            tmp += E[1][1]*e_du<1,1,p,q>(u);
            
            return a*trE_dupq + 2*b*trE*trE_dupq + 2*c*tmp + d*dGamma*det_du_tmpl<p,q>(u);
        }
        
        /** \brief Compute the first derivative of the Ogden energy functional 
            \param d The Ogden constant
            \param lambda, my The material parameters
        */
        static void Ogden_du_tmpl(const FieldMatrix<double,3,3>& u, 
                                  double d, double lambda, double my, FieldMatrix<double,3,3>& fu) {
            double a = -d*Gamma_x(1);
            double b = (lambda - d*(Gamma_x(1)+Gamma_xx(1)))/2;
            double c = my + d*Gamma_x(1);
            
            double E[3][3];
            for (int i=0; i<3; ++i)
                for (int j=0; j<3; ++j)
                    E[i][j] = E_val(u,i,j);
            
            // Compute tr(E)
            double trE = E[0][0] + E[1][1] + E[2][2];
            
            double dGamma = Gamma_x(det_val(u));
            
            // Collect results
            fu[0][0] = compute_fu<0,0>(u,E,trE,dGamma,a,b,c,d);
            fu[0][1] = compute_fu<0,1>(u,E,trE,dGamma,a,b,c,d);
            fu[0][2] = compute_fu<0,2>(u,E,trE,dGamma,a,b,c,d);
            fu[1][0] = compute_fu<1,0>(u,E,trE,dGamma,a,b,c,d);
            fu[1][1] = compute_fu<1,1>(u,E,trE,dGamma,a,b,c,d);
            fu[1][2] = compute_fu<1,2>(u,E,trE,dGamma,a,b,c,d);
            fu[2][0] = compute_fu<2,0>(u,E,trE,dGamma,a,b,c,d);
            fu[2][1] = compute_fu<2,1>(u,E,trE,dGamma,a,b,c,d);
            fu[2][2] = compute_fu<2,2>(u,E,trE,dGamma,a,b,c,d);
            
        }

        static void Ogden_du_tmpl(const FieldMatrix<double,2,2>& u, 
                                  double d, double lambda, double my, 
                                  FieldMatrix<double,2,2>& fu) {
            double a = -d*Gamma_x(1);
            double b = (lambda - d*(Gamma_x(1)+Gamma_xx(1)))/2;
            double c = my + d*Gamma_x(1);
            
            double E[2][2];
            for (int i=0; i<2; ++i)
                for (int j=0; j<2; ++j)
                    E[i][j] = E_val(u,i,j);
            
            // Compute tr(E)
            double trE = E[0][0] + E[1][1];
            
            double dGamma = Gamma_x(det_val(u));
            
            // Collect results
            fu[0][0] = compute_fu<0,0>(u,E,trE,dGamma,a,b,c,d);
            fu[0][1] = compute_fu<0,1>(u,E,trE,dGamma,a,b,c,d);
            fu[1][0] = compute_fu<1,0>(u,E,trE,dGamma,a,b,c,d);
            fu[1][1] = compute_fu<1,1>(u,E,trE,dGamma,a,b,c,d);
            
        }
        
        /*----------------------------------------------------------------------------*/
        
        template <int p, int q, int r, int s>
        static double compute_fuu_1(const FieldMatrix<double,3,3>& u, 
                                    double const E[3][3], double trE, double const trE_du[3][3],
                                    double detu, double dGamma, double ddGamma, 
                                    double a, double b, double c, double d) {
            double trE_dupqrs = e_dudu<0,0,p,q,r,s>()+e_dudu<1,1,p,q,r,s>()+e_dudu<2,2,p,q,r,s>();
            
            double result = a*trE_dupqrs + 2*b*( trE*trE_dupqrs + trE_du[r][s]*trE_du[p][q] );
            
            double tmp = 0;
            tmp += e_du<0,0,r,s>(u)*e_du<0,0,p,q>(u) + E[0][0]*e_dudu<0,0,p,q,r,s>();
            tmp += e_du<0,1,r,s>(u)*e_du<0,1,p,q>(u) + E[0][1]*e_dudu<0,1,p,q,r,s>();
            tmp += e_du<0,2,r,s>(u)*e_du<0,2,p,q>(u) + E[0][2]*e_dudu<0,2,p,q,r,s>();
            tmp += e_du<1,0,r,s>(u)*e_du<1,0,p,q>(u) + E[1][0]*e_dudu<1,0,p,q,r,s>();
            tmp += e_du<1,1,r,s>(u)*e_du<1,1,p,q>(u) + E[1][1]*e_dudu<1,1,p,q,r,s>();
            tmp += e_du<1,2,r,s>(u)*e_du<1,2,p,q>(u) + E[1][2]*e_dudu<1,2,p,q,r,s>();
            tmp += e_du<2,0,r,s>(u)*e_du<2,0,p,q>(u) + E[2][0]*e_dudu<2,0,p,q,r,s>();
            tmp += e_du<2,1,r,s>(u)*e_du<2,1,p,q>(u) + E[2][1]*e_dudu<2,1,p,q,r,s>();
            tmp += e_du<2,2,r,s>(u)*e_du<2,2,p,q>(u) + E[2][2]*e_dudu<2,2,p,q,r,s>();
            
            result += 2*c*tmp + d*( dGamma*det_dudu_tmpl<p,q,r,s>(u)
                                    + ddGamma*det_du_tmpl<p,q>(u)*det_du_tmpl<r,s>(u) );
            // precomputating and storing det_du(r,s) seems to yield exactly the same performance...
            
            
            return result;
        }

        template <int p, int q, int r, int s>
        static double compute_fuu_1(const FieldMatrix<double,2,2>& u, 
                                    double const E[2][2], double trE, double const trE_du[2][2],
                                    double detu, double dGamma, double ddGamma, 
                                    double a, double b, double c, double d) {
            double trE_dupqrs = e_dudu<0,0,p,q,r,s>()+e_dudu<1,1,p,q,r,s>();
            
            double result = a*trE_dupqrs + 2*b*( trE*trE_dupqrs + trE_du[r][s]*trE_du[p][q] );
            
            double tmp = 0;
            tmp += e_du<0,0,r,s>(u)*e_du<0,0,p,q>(u) + E[0][0]*e_dudu<0,0,p,q,r,s>();
            tmp += e_du<0,1,r,s>(u)*e_du<0,1,p,q>(u) + E[0][1]*e_dudu<0,1,p,q,r,s>();
            tmp += e_du<1,0,r,s>(u)*e_du<1,0,p,q>(u) + E[1][0]*e_dudu<1,0,p,q,r,s>();
            tmp += e_du<1,1,r,s>(u)*e_du<1,1,p,q>(u) + E[1][1]*e_dudu<1,1,p,q,r,s>();
            
            result += 2*c*tmp + d*( dGamma*det_dudu_tmpl<p,q,r,s>(u)
                                    + ddGamma*det_du_tmpl<p,q>(u)*det_du_tmpl<r,s>(u) );
            // precomputating and storing det_du(r,s) seems to yield exactly the same performance...
            
            
            return result;
        }
        
        template <int p, int q>
        static void compute_fuu(const FieldMatrix<double,3,3>& u, 
                                double const E[3][3], double trE, double const trE_du[3][3],
                                double detu, double dGamma, double ddGamma, 
                                double a, double b, double c, double d,
                                double fuu[3][3]) {
            fuu[0][0] = compute_fuu_1<p,q,0,0>(u,E,trE,trE_du,detu,dGamma,ddGamma,a,b,c,d);
            fuu[0][1] = compute_fuu_1<p,q,0,1>(u,E,trE,trE_du,detu,dGamma,ddGamma,a,b,c,d);
            fuu[0][2] = compute_fuu_1<p,q,0,2>(u,E,trE,trE_du,detu,dGamma,ddGamma,a,b,c,d);
            fuu[1][0] = compute_fuu_1<p,q,1,0>(u,E,trE,trE_du,detu,dGamma,ddGamma,a,b,c,d);
            fuu[1][1] = compute_fuu_1<p,q,1,1>(u,E,trE,trE_du,detu,dGamma,ddGamma,a,b,c,d);
            fuu[1][2] = compute_fuu_1<p,q,1,2>(u,E,trE,trE_du,detu,dGamma,ddGamma,a,b,c,d);
            fuu[2][0] = compute_fuu_1<p,q,2,0>(u,E,trE,trE_du,detu,dGamma,ddGamma,a,b,c,d);
            fuu[2][1] = compute_fuu_1<p,q,2,1>(u,E,trE,trE_du,detu,dGamma,ddGamma,a,b,c,d);
            fuu[2][2] = compute_fuu_1<p,q,2,2>(u,E,trE,trE_du,detu,dGamma,ddGamma,a,b,c,d);
        }

        template <int p, int q>
        static void compute_fuu(const FieldMatrix<double,2,2>& u, 
                                double const E[2][2], double trE, double const trE_du[2][2],
                                double detu, double dGamma, double ddGamma, 
                                double a, double b, double c, double d,
                                double fuu[2][2]) {
            fuu[0][0] = compute_fuu_1<p,q,0,0>(u,E,trE,trE_du,detu,dGamma,ddGamma,a,b,c,d);
            fuu[0][1] = compute_fuu_1<p,q,0,1>(u,E,trE,trE_du,detu,dGamma,ddGamma,a,b,c,d);
            fuu[1][0] = compute_fuu_1<p,q,1,0>(u,E,trE,trE_du,detu,dGamma,ddGamma,a,b,c,d);
            fuu[1][1] = compute_fuu_1<p,q,1,1>(u,E,trE,trE_du,detu,dGamma,ddGamma,a,b,c,d);
        }
        
        
        static void Ogden_dudu_tmpl(const FieldMatrix<double, 3,3>& u, 
                                    double d, double lambda, double my, double fuu[3][3][3][3]) {
            double a = -d*Gamma_x(1);
            double b = (lambda - d*(Gamma_x(1)+Gamma_xx(1)))/2;
            double c = my + d*Gamma_x(1);
            
            double E[3][3];
            for (int i=0; i<3; ++i)
                for (int j=0; j<3; ++j)
                    E[i][j] = E_val(u,i,j);
            
            // Compute tr(E)
            double trE = E[0][0] + E[1][1] + E[2][2];
            
            double trE_du[3][3];
            for (int i=0; i<3; ++i)
                for (int j=0; j<3; ++j)
                    trE_du[i][j] = E_du(u,0,0,i,j)+E_du(u,1,1,i,j)+E_du(u,2,2,i,j);
            
            double detu = det_val(u);
            double dGamma = Gamma_x(detu);
            double ddGamma = Gamma_xx(detu);
            
            
            // Collect results
            compute_fuu<0,0>(u,E,trE,trE_du,detu,dGamma,ddGamma,a,b,c,d,fuu[0][0]);
            compute_fuu<0,1>(u,E,trE,trE_du,detu,dGamma,ddGamma,a,b,c,d,fuu[0][1]);
            compute_fuu<0,2>(u,E,trE,trE_du,detu,dGamma,ddGamma,a,b,c,d,fuu[0][2]);
            compute_fuu<1,0>(u,E,trE,trE_du,detu,dGamma,ddGamma,a,b,c,d,fuu[1][0]);
            compute_fuu<1,1>(u,E,trE,trE_du,detu,dGamma,ddGamma,a,b,c,d,fuu[1][1]);
            compute_fuu<1,2>(u,E,trE,trE_du,detu,dGamma,ddGamma,a,b,c,d,fuu[1][2]);
            compute_fuu<2,0>(u,E,trE,trE_du,detu,dGamma,ddGamma,a,b,c,d,fuu[2][0]);
            compute_fuu<2,1>(u,E,trE,trE_du,detu,dGamma,ddGamma,a,b,c,d,fuu[2][1]);
            compute_fuu<2,2>(u,E,trE,trE_du,detu,dGamma,ddGamma,a,b,c,d,fuu[2][2]);
        }

        static void Ogden_dudu_tmpl(const FieldMatrix<double, 2,2>& u, 
                                    double d, double lambda, double my, double fuu[2][2][2][2]) {
            double a = -d*Gamma_x(1);
            double b = (lambda - d*(Gamma_x(1)+Gamma_xx(1)))/2;
            double c = my + d*Gamma_x(1);
            
            double E[2][2];
            for (int i=0; i<2; ++i)
                for (int j=0; j<2; ++j)
                    E[i][j] = E_val(u,i,j);
            
            // Compute tr(E)
            double trE = E[0][0] + E[1][1];
            
            double trE_du[2][2];
            for (int i=0; i<2; ++i)
                for (int j=0; j<2; ++j)
                    trE_du[i][j] = E_du(u,0,0,i,j)+E_du(u,1,1,i,j);
            
            double detu = det_val(u);
            double dGamma = Gamma_x(detu);
            double ddGamma = Gamma_xx(detu);
            
            // Collect results
            compute_fuu<0,0>(u,E,trE,trE_du,detu,dGamma,ddGamma,a,b,c,d,fuu[0][0]);
            compute_fuu<0,1>(u,E,trE,trE_du,detu,dGamma,ddGamma,a,b,c,d,fuu[0][1]);
            compute_fuu<1,0>(u,E,trE,trE_du,detu,dGamma,ddGamma,a,b,c,d,fuu[1][0]);
            compute_fuu<1,1>(u,E,trE,trE_du,detu,dGamma,ddGamma,a,b,c,d,fuu[1][1]);
        }
        
        /*----------------------------------------------------------------------------*/
        
        
    }; // end class
    
} // end namespace 

#include "ogdenassembler.cc"

#endif

